/************************************************************

	tuner.c

	control for "SANYO AM/FM PLL Frequency Synthesizer LC72131"
	using the SANYO CCB(computer control bus) audio LSI serial buf format.

	(c) 2013/08/13 Suwa   http://www.suwa-koubou.jp

  2013/08/29 fix PLL_DATA_IN()
************************************************************/

#include <stdio.h>
#include "Compiler.h"
#include "HardwareProfile.h"
#include "GenericTypeDefs.h"
#include "tuner.h"
#include "pic_delay.h"
#include "uart2.h"

/*-----------------------------------------------------------
	definition
-----------------------------------------------------------*/
// test circuit
//#define PLL_CE_ON()       LATDSET=0x0002  // RD1 ON
//#define PLL_CE_OFF()      LATDCLR=0x0002  // RD1 OFF
//#define PLL_DATA_ON()     LATDSET=0x0004  // RD2 ON
//#define PLL_DATA_OFF()    LATDCLR=0x0004  // RD2 OFF
//#define PLL_CLOCK_ON()	LATDSET=0x0008  // RD3 ON
//#define PLL_CLOCK_OFF()   LATDCLR=0x0008  // RD3 OFF
//#define PLL_WAIT()        delay_5us()  
//#define PLL_DATA_IN()     (PORTD & 0x10)  // RD4 input

// release circuit
#define PLL_CE_ON()       LATGSET=0x0040  // RG6 ON
#define PLL_CE_OFF()      LATGCLR=0x0040  // RG6 OFF
#define PLL_DATA_ON()     LATGSET=0x0080  // RG7 ON
#define PLL_DATA_OFF()    LATGCLR=0x0080  // RG7 OFF
#define	PLL_CLOCK_ON()	  LATGSET=0x0100  // RG8 ON
#define PLL_CLOCK_OFF()   LATGCLR=0x0100  // RG8 OFF
#define PLL_WAIT()        delay_5us()  
//#define PLL_DATA_IN()     (PORTD & 0x0200)  // RG9 input
#define PLL_DATA_IN()     (PORTG & 0x0200)  // RG9 input  fix 2013/08/29


#define FM_IF_min  681600    // 10.7MHz-50KHz = 10,650,000 * 0.064 = 681,600
#define	FM_IF_max  688000    // 10.7Mhz+50KHz = 10,750,000 * 0.064 = 688,000
#define AM_IF_min   28160    // 450KHz-10KHz = 440,000 * 0.064 = 28,160
#define AM_IF_max   29440    // 450KHz+10KHz = 460,000 * 0.064 = 29,440

/*-----------------------------------------------------------
	function prototype
-----------------------------------------------------------*/

static	void    pllIN1Out(void);
static	void    pllIN2Out(void);
static	void    pllOUTInput(void);
static	void    pllDataOut(BYTE d);
static	BYTE    pllDataInput(void);


/*-----------------------------------------------------------
	global variable
-----------------------------------------------------------*/
// IN1
static WORD P_CTR;
static BYTE SNS, DVS, CTE, XS, R_CTR;

// IN2
static BYTE IO_C, IO, BO, DO_C, UL, DZ_C, GT;
static BYTE TBC, DLC, IFS, TEST;

// OUT
static DWORD IF_CTR;
// BYTE IO, UL share IN2 

/*==========================================================
	Tuned xxxKHz
	AM 530KHz --> 1610KHz
    FM 76.0MHz --> 76000,  90.0MHz --> 90000
===========================================================*/

int Tuned(DWORD freq)
{
//	char pbuf[32];
	int	tune_trys;
	DWORD IF_min, IF_max;


	// valid frequency ?
	if(freq < 530 || (freq > 1610 && freq < 76000) || freq > 90000){
		return TUNE_FAIL;
	}

	// AM ?
	if(freq <= 1610){
		// set AM
		DVS = 0; SNS = 0;
		// IF pll divider
		R_CTR = 9; // 9Khz
		P_CTR = ((freq + 450)/9) << 4;	// don't care P0-P3, LSB is P4 
		IF_min = AM_IF_min;
		IF_max = AM_IF_max;
	}else{
		// set FM
		DVS = 1, SNS = 0;
		// IF pll divider
		R_CTR = 3;	// 25KHz
		P_CTR = (freq - 10700)/(25*2);
		IF_min = FM_IF_min;
		IF_max = FM_IF_max;
	}

	// select Xtal. frequency
	XS = 1;		// 7.2MHz		

	// set IO pin direction
	IO_C = 3;	// IO1, IO2 are output.
	GT = 3;     // IF count time 64msec
	DO_C = 0;	// DO pin are open.
	UL = 0;		// open
	DZ_C = 0;	// select DZA
	TBC = 0;	// do not use clock time base
	DLC = 0;	// normal charge pomp out
	IFS = 1;	// normal set 1
	TEST = 0;	// default value 

	// FM sequense
	// 1st. CTE = 0, IO = 2, BO = 14;
	// 2nd. CTE = 1, IO = 0, BO = 14;
	// 3rd. CTE = 0, IO = 2, BO = 10;

	// 1st.
	if(freq <= 1610) BO = 7; // AM mono
	else BO = 14; // FM mono
	CTE = 0, IO = 2;
	pllIN1Out();
	delay_10us(5);
	pllIN2Out();
	delay_10us(5);

	// 2nd. check SD, IF
	for(IF_CTR = 0, tune_trys = 3; tune_trys; tune_trys--){ 
		if(freq <= 1610) BO = 7; // AM mono
		else BO = 14; // FM mono
		CTE = 1, IO = 0;
		pllIN1Out();
		delay_10us(5);
		pllIN2Out();
		delay_10us(5);
		delay_ms(100);	// over 64msec
		pllOUTInput();
		delay_10us(5);

		// sprintf(pbuf, "IF_CTR=%lu\r\n", IF_CTR); putstr(pbuf);
		// sprintf(pbuf, "UL=%d\r\n", UL ? 1: 0); putstr(pbuf);

		// tuning ok?
		if(IF_CTR > IF_min && IF_CTR < IF_max){
			break;
		}
	}

	// 3rd. is FM to STEREO
	if(freq >= 76000){
		CTE = 0, IO = 2, BO = 10; // stereo
		UL = 0;
		pllIN1Out();
		delay_10us(5);
		pllIN2Out();
		delay_10us(5);
	}

	if(tune_trys == 0) return TUNE_FAIL;
	return TUNE_SUCCESS;
}


/*-----------------------------------------------------------
	IN1 DataOut
-----------------------------------------------------------*/
static
void pllIN1Out(void)
{
	BYTE Data11, Data12, Data13;

	Data11 = P_CTR & 0xff;
	Data12 = (P_CTR >> 8) & 0xff;
	Data13 = (SNS & 1);
	Data13 |= ((DVS & 1) << 1);
	Data13 |= ((CTE & 1) << 2);
	Data13 |= ((XS & 1) << 3);
	Data13 |= ((R_CTR & 0x0f) << 4);

	pllDataOut(0x28);
	PLL_WAIT();
	PLL_CE_ON();
	pllDataOut(Data11);
	pllDataOut(Data12);
	pllDataOut(Data13);
	PLL_CE_OFF();

}


/*-----------------------------------------------------------
	IN2 DataOut
-----------------------------------------------------------*/
static
void pllIN2Out(void)
{
	BYTE Data21, Data22, Data23;

	Data21 = IO_C & 3;
	Data21 |= ((IO & 3) << 2);
	Data21 |= ((BO & 0x0f) << 4);

	Data22 = ((DO_C & 0x07) << 1);
	Data22 |= ((UL & 3) << 4);
	Data22 |= ((DZ_C & 3) << 6);

	Data23 = GT & 3;
	Data23 |= ((TBC & 1) << 2);
	Data23 |= ((DLC & 1) << 3);
	Data23 |= ((IFS & 1) << 4);
	Data23 |= ((TEST & 0x07) << 5);

	pllDataOut(0x29);
	PLL_WAIT();
	PLL_CE_ON();
	pllDataOut(Data21);
	pllDataOut(Data22);
	pllDataOut(Data23);
	PLL_CE_OFF();

}


/*-----------------------------------------------------------
	OUT DataInput
-----------------------------------------------------------*/
static
void pllOUTInput(void)
{
	BYTE Data31, Data32, Data33;

	pllDataOut(0x2a);
	PLL_WAIT();
	PLL_CE_ON();
	Data31 = pllDataInput();
	Data32 = pllDataInput();
	Data33 = pllDataInput();
	PLL_CE_OFF();

	IO = (Data31 & 0xc0) >> 6;
	UL = (Data31 & 0x10) >> 4;

	IF_CTR = (((DWORD)Data31 & 0x0f) << 16)| ((DWORD)Data32 << 8) | (DWORD)Data33;

}


/*-----------------------------------------------------------
	pllDataOut()
-----------------------------------------------------------*/
static
void pllDataOut(BYTE d)
{
	int n;

	for(n = 0; n < 8; n++){
		if(d & 1){
			PLL_DATA_ON();
		}else{
			PLL_DATA_OFF();
		}
		PLL_WAIT();
		PLL_CLOCK_ON();
		PLL_WAIT();
		PLL_CLOCK_OFF();
		d >>= 1;
	}
	PLL_DATA_OFF();
}	


/*-----------------------------------------------------------
	pllDataInput()
-----------------------------------------------------------*/
static
BYTE pllDataInput(void)
{
	BYTE d;
	int n;

	d = 0;
	for(n = 0; n < 8; n++){
		d <<= 1;
		PLL_WAIT();
		PLL_CLOCK_ON();
		if(PLL_DATA_IN()){
			d |= 1;
		}
		PLL_WAIT();
		PLL_CLOCK_OFF();
	}
	return d;
}


/*** end of tuner.c ****************************************/

